From iris.proofmode Require Import base coq_tactics reduction tactics.
From diaframe Require Import env_utils util_classes solve_defs.
From diaframe.steps Require Import namer_tacs pure_solver solve_sep transform introduce_hyp introduce_var solve_one 
      solve_and solve_sep_foc solve_one_foc disj_chooser_tacs.
From iris.bi Require Import bi telescopes.


(* Apart from iStep, Diaframe also defines a (more fine-grained (😉) ) solveStep tactic,
   which executes precisely one instruction of Diaframe's proof search strategy.

   If a hint you expect to be applied is not applied or found, this file comes with a 'solveStep with H' tactic.
   The next goals will be:
   - prove a hint from hypothesis H to the current focused atom
   - continuation goal after applying this hint.
*)

(* We need to do some juggling with evars, which the next few tactics are for *)
Ltac hide_evars_adv next_tac name_hook := 
  match goal with
  | |- context [?arg] =>
    is_evar arg;
    let evar_name := fresh "x" in
    set (evar_name := arg);
    try (hide_evars_adv next_tac name_hook); 
    (* it is important that failure of next_tac will not cause evars to be unfolded *)
    name_hook evar_name
  | |- _ => next_tac
  end.

Ltac unhider evar_name :=
  unfold evar_name; clear evar_name.

Ltac noop_tac evar_name := idtac.

Ltac hide_evars_then next_tac :=
  hide_evars_adv next_tac unhider.

Tactic Notation "hide_evars" "then" tactic(next) "then" "unhide" :=
  hide_evars_then next.

Tactic Notation "hide_evars" := hide_evars_adv idtac noop_tac.


Lemma tac_pure_intro_direct `(Δ : envs PROP) (φ : Prop) :
  φ → envs_entails Δ ⌜φ⌝.
Proof. intros. eapply tac_pure_intro; [tc_solve| split| exact H]. Qed.

Ltac maybeIPureIntro :=
  lazymatch goal with
  | |- envs_entails _ _ => 
      refine (tac_pure_intro_direct _ _ _)
  | |- _ => idtac
  end.

(* Now, we define the tactic that tries to solve pure goals.
   It fails if the pure goal has to be solved now, but can't (since some evar needs to be instantiated).
   It shelves the pure goal if it cant be solved and doesnt have to be solved now *)
Ltac trySolvePurePreSplit Δφ R :=
  (* fancy pure stuff from RefinedC should go here *)
  let φ := 
    lazymatch Δφ with
    | envs_entails _ ⌜?φ⌝ => φ
    | _ => Δφ
    end in
  let φr :=
    lazymatch φ with
    | BehindModal _ ?φr => φr
    | _ => φ
    end in
  first (* check if we are not already unprovable. If we are, fail hard. *)
  [ match φ with
    | @eq _ ?a ?b => is_evar a; fail 1 "dont try negating evar equalities"
    | _ => idtac
    end; 
    assert (¬ φr) as _ by (hide_evars then (idtac; trySolvePure) then unhide); 
    (* if idtac is not present, trySolvePure will run immediately...? *)
    fail 1 "Current pure goal is provably untrue! Something went terribly wrong!"
  | has_evar φ; 
    ((split; first (solve [maybeIPureIntro; trySolvePure])) ||
    (* if there is an evar, we need to solve it now *)
    (* somehow prepending idtac changes the semantics and makes it work like intentioned...?*)
    (idtac; fail 1 "Pure condition"φ"needs to be proved now!"))
  (* final option: solve it now, and if unsuccesful, shelve it *)
   | split; [first [solve [maybeIPureIntro; trySolvePure] | (change φ with φr; shelve)] | ] ].

Ltac introStepNT namer_tac := (* used for iDecompose tactic *)
  (simplify_eq; subst) ||
  lazymatch goal with
  | |- envs_entails _ ?R =>
    lazymatch R with
    | (IntroduceVars _) => introduceVarStep
    | (IntroduceHyp _ _) => introduceHypStep namer_tac
    end
  end.

Ltac solveStepNT namer_tac chooser_tac :=
  (simplify_eq; subst) ||
  lazymatch goal with
  | |- envs_entails _ ?R =>
    lazymatch R with
    | (SolveSep _ _ _) => solveSepStep namer_tac
    | (Transform _) => transformStep namer_tac
    | (IntroduceVars _) => introduceVarStep
    | (IntroduceHyp _ _) => introduceHypStep namer_tac
    | (SolveOne _ _) => solveOneStep namer_tac
    | (SolveAnd _ _) => solveAndStep
    | (SolveSepFoc _ _ _ _) => solveSepFocStep namer_tac chooser_tac
    | (SolveOneFoc _ _ _) => solveOneFocStep namer_tac chooser_tac
    end
  | |- ex _ => eexists
  | |- ?φ ∧ ?R => 
    trySolvePurePreSplit φ R
  | |- ?φ =>
    solve [trySolvePure]
  end.

Ltac solveStepsNT namer_tac chooser_tac :=
  first [solveStepNT namer_tac chooser_tac; solveStepsNT namer_tac chooser_tac | idtac].


Ltac solveStep_with mi post_tac := 
  lazymatch goal with
  | |- environments.envs_entails _ (SolveSep _ _ _) =>
    notypeclasses refine (tac_solve_sep_ext_biabd_disj _ mi _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _);
        [tc_solve | post_tac () | idtac..]
  | |- environments.envs_entails _ (SolveOne _ _) =>
    notypeclasses refine (solve_one.tac_solve_one_ext_abd_input _ mi _ _ _ _ _ _ _); 
      [post_tac () | idtac..]
  | |- environments.envs_entails _ (SolveSepFoc _ _ _ _) =>
    notypeclasses refine (tac_solve_sep_foc_biabd_disj _ mi _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _);
      [tc_solve | post_tac () | idtac..]
  | |- environments.envs_entails _ (SolveOneFoc _ _ _) => 
    notypeclasses refine (tac_solve_one_foc_abduct_foc_disj_ex _ mi _ _ _ _ _ _ _ _); 
      [post_tac () | idtac ]
  end.

Tactic Notation "solveStep" "with" constr(name) :=
  let mi := constr:(Some $ INamed name) in
  let post_tac _ := eapply fic_ext_lookup_Some; first reflexivity in
  solveStep_with mi post_tac.

Tactic Notation "solveStep" "with" "empty_hyp_first" :=
  let mi := constr:(@None ident) in
  let post_tac _ := (eapply fic_ext_lookup_empty_first) in
  solveStep_with mi post_tac.

Tactic Notation "solveStep" "with" "ε₀" := solveStep with empty_hyp_first.

Tactic Notation "solveStep" "with" "empty_hyp_last" :=
  let mi := constr:(@None ident) in
  let post_tac _ := (eapply fic_ext_lookup_empty_last) in
  solveStep_with mi post_tac.

Tactic Notation "solveStep" "with" "ε₁" := solveStep with empty_hyp_last.

Ltac solveStepDT delete_tac := 
  let namer := ltac:(default_namer delete_tac) in
  solveStepNT namer chooser_always_left.

Ltac solveStep := solveStepDT default_delete_tac.

Ltac solveStepsDT delete_tac := 
  let namer := ltac:(default_namer delete_tac) in
  solveStepsNT namer chooser_always_left.

Ltac solveSteps := solveStepsDT default_delete_tac.

Ltac introStepDT delete_tac := introStepNT ltac:(default_namer delete_tac).

Ltac introStep := introStepDT default_delete_tac.
