From Ltac2 Require Export Ltac2 Printf.
From Coq.micromega Require Import Psatz Zify.

Export Set Default Proof Mode "Classic".


(** This file provides a way to detect the newly introduced variables
   by any tactic. 

   This approach does nót work if the tactic itself does something like
     [revert H; move => H]: our tactic will notice that H got removed somewhere
   in the meantime, so will require you to rename it again.

   The [zify] tactic used by [lia] does this. This file overrides
   [Zify.zify_pre_hook] to fix that.
*)

Inductive RecordCurrentVariables : Prop := recorded.

Ltac2 get_idents (u : unit) : ident list :=
  List.map (fun (ident, _, _) => ident) (Control.hyps ()).

Ltac2 cbv_flags : Std.red_flags := {
  Std.rBeta := true;
  Std.rMatch := true;
  Std.rFix := true;
  Std.rCofix := true;
  Std.rZeta := true;
  Std.rDelta := true;
  Std.rConst := [];
}.

Ltac2 is_solved_goal (_ : unit) :=
  let cell : bool ref := { contents := false } in
  Control.enter (fun _ => cell.(contents) := true);
  Bool.neg (cell.(contents)).

Ltac2 count_goals (_ : unit) :=
  let cnt : int ref := { contents := 0} in
  Control.enter (fun _ =>
    (* TODO: is this run concurrently? *)
    cnt.(contents) := Int.add (cnt.(contents)) 1
  );
  cnt.(contents).

Ltac2 focus_last (continuation : unit -> 'a) : 'a :=
  let n := count_goals () in
  Control.focus n n continuation.

Ltac2 find_fresh_idents_between (u : unit) : (unit -> ident list) * (unit -> unit) :=
  (** If this function is called, it returns two closure. If the first
     closure is called on a single goal, it will return all the idents that have
     been newly introduced into the Coq context. Once you are done,
     you are required to call the second closure to clean up *)

  (** Unfortunately, doing that is a bit more complicated than looking 
    at the difference between two calls to [Control.hyps ()]. That approach
    will not notice if an ident used in the old context got cleared, and
    a new ident (of possibly different type!) got introduced with the same name.
    To do this, we piggyback on the way evars keep an account of which variables
    were in scope when they got created. *)
  let fresh_evar := (* creates a fresh evar of type [RecordCurrentVariables] *)
    let x := '(_ : RecordCurrentVariables) in
    match Constr.Unsafe.kind x with
    | Constr.Unsafe.Cast c _ _ => 
      c
    | _ => Control.throw Not_found
    end in 

  let checkme := Fresh.in_goal (Option.get (Ident.of_string "checkme")) in
  (* create a _fresh_ name which we will use to reference this evar in our goal *)
  Std.pose (Some checkme) fresh_evar;
  let checkme_constr := (Constr.Unsafe.make (Constr.Unsafe.Var checkme)) in
  (* and 'entangles' it with the current proof goal. This is crucial for 
    getting Coq to notice that it needs to start enforcing constraints.

    We precompute the idents present in the constraints of [checkme] *)
  let (old_idents, raw_evar) :=
    let raw_evar := Std.eval_cbv cbv_flags checkme_constr in
    match Constr.Unsafe.kind raw_evar with
    | Constr.Unsafe.Evar e cs =>
      (Array.to_list (Array.map (fun c =>
        match Constr.Unsafe.kind c with
        | Constr.Unsafe.Var i => i
        | _ => Control.throw Not_found
        end
      ) cs), e)
    | _ => Control.throw Not_found
    end in

  (* In case all goals get solved, we want to prevent giving the user a 'shelved goal'.
     The client should manually check if all goals get solved and call the below cleanup
     closure to get rid of the uninstantiated evar. *)
  let cleanup := (fun _ => 
    Control.new_goal raw_evar; 
    focus_last (fun _ => exact recorded)
  ) in

  (* and return the closure: *)
  ((fun () => 
    (* We need to call [cbv] to get rid of the let-binding, _and_ to 
      refresh the constraints. I am not sure why the constraint refreshing
      only happens after [cbv] *)
    let refreshed_evar := Std.eval_cbv cbv_flags checkme_constr in

    (* We jump through some hoops to get out precisely the constraints
      that are idents ([Constr.Unsafe.Var]). *)
    let new_constrs :=
      match Constr.Unsafe.kind refreshed_evar with
      | Constr.Unsafe.Evar e cs =>
        Array.to_list cs
      | _ => Control.throw Not_found
      end in
    let new_var_constrs_only := List.filter Constr.is_var new_constrs in
    let new_idents_only := List.map (fun c => 
      match Constr.Unsafe.kind c with
      | Constr.Unsafe.Var i => i
      | _ => Control.throw Not_found
      end) new_var_constrs_only in

    (* Then, we check which of those idents were already present in [old_idents].
      These are the idents which have been 'kept' since this closure was generated. *)
    let kept_idents := List.filter (fun i => List.mem Ident.equal i old_idents) new_idents_only in

    (* We now have all the information we need. We clear (but do not unify!) the evar from this goal. *)
    Std.clear [checkme];
    (* We do not unify so that other goals can still read the required information from the evar. *)

    (* We return precisely the idents which are not in [kept_idents] *)
    List.filter (fun i => Bool.neg (List.mem Ident.equal i kept_idents)) (get_idents ())
  ), cleanup).

Ltac2 rec zip (ls1 : 'a list) (ls2 : 'b list) :=
  match ls1 with
  | [] => []
  | a :: ls1 =>
    match ls2 with
    | [] => []
    | b :: ls2 =>
      (a, b) :: zip ls1 ls2
    end
  end.

Ltac2 rec firstn_safe (n : int) (ls : 'a list) :=
  Control.assert_valid_argument "firstn_safe" (Int.ge n 0);
  match Int.equal n 0 with
  | true => []
  | false
    => match ls with
       | [] => []
       | x :: xs
         => x :: firstn_safe (Int.sub n 1) xs
       end
  end.

Ltac2 rec skipn_safe (n : int) (ls : 'a list) :=
  Control.assert_valid_argument "skipn_safe" (Int.ge n 0);
  match Int.equal n 0 with
  | true => ls
  | false
    => match ls with
       | [] => []
       | x :: xs
         => skipn_safe (Int.sub n 1) xs
       end
  end.

Ltac2 find_dependent_on_ident (i : ident) : ident :=
  multi_match! goal with
  | [ hyp : context [?i_constr] |- _] =>
    Std.unify i_constr (Control.hyp i);
    hyp
  end.

Ltac2 Type exn ::= [MoreValues].

Ltac2 Type 'a set := { mutable set_as_list : 'a list; el_equal : 'a -> 'a -> bool }.

Ltac2 empty_set (eq : 'a -> 'a -> bool) := { set_as_list := []; el_equal := eq}.

Ltac2 set_add (a_set : 'a set) (el : 'a) : unit :=
  if List.mem (a_set.(el_equal)) el (a_set.(set_as_list)) then
    ()
  else
    a_set.(set_as_list) := el :: a_set.(set_as_list).

Ltac2 set_remove (a_set : 'a set) (el : 'a) : unit :=
  let rec new_list (tail : 'a list) :=
    match tail with
    | [] => []
    | hd :: tail' =>
      if a_set.(el_equal) hd el then
        tail'
      else
        hd :: new_list tail'
    end
  in
  a_set.(set_as_list) := new_list (a_set.(set_as_list)).

Ltac2 set_from_list (els : 'a list) (eq : 'a -> 'a -> bool) :=
  let result : 'a set := empty_set eq in
  List.iter (set_add result) els;
  result.

Ltac2 rec get_all (closure : unit -> 'a) : 'a list :=
  match Control.case closure with
  | Val pr =>
    let (hd, tailgen) := pr in
    hd :: get_all (fun _ => tailgen MoreValues)
  | Err exn =>
    []
  end.

Ltac2 add_all_to_set (closure : unit -> 'a) (result : 'a set) : unit :=
  let rec get_more (c : unit -> 'a) : unit :=
    match Control.case c with
    | Val pr =>
      let (hd, tailgen) := pr in
      set_add result hd;
      get_more (fun _ => tailgen MoreValues)
    | Err exn =>
      ()
    end in
  get_more closure.

Ltac2 find_all_dependent_on_ident (i : ident) : ident set :=
  let result := empty_set Ident.equal in
  add_all_to_set (fun _ => find_dependent_on_ident i) result;
  result.

Ltac2 find_all_dependent_on_idents (is : ident list) : ident list :=
  let result := empty_set Ident.equal in
  List.iter (fun i => add_all_to_set (fun _ => find_dependent_on_ident i) result) is;
  List.iter (fun i => set_remove result i) is;
  result.(set_as_list).


Ltac2 to_list_of_size 
  (els : 'a list) (size : int) (default : 'a)
  (too_short_trig : int -> unit) (too_long_trig : int -> unit) : 'a list :=
  let els_len := List.length els in
  if Int.lt els_len size then
    too_short_trig (Int.sub size els_len);
    List.append els (List.repeat default (Int.sub size els_len))
  else
    if Int.gt els_len size then
      (* silently discard stuff if no goals remain. *)
      if Int.equal size 0 then () else too_long_trig (Int.sub els_len size);
      List.firstn size els
    else
      els.


Ltac2 Type ReintroductionWarningBehavior := [ Silent | Verbose ].

Ltac2 Type NamingProblems := {
  mutable global_problems : (unit -> unit) option;
  mutable subgoal_problems : (unit -> unit) option list;
}.

Ltac2 maybe_closure_append (omsg1 : (unit -> unit) option) (msg2 : unit -> unit) : unit -> unit :=
  match omsg1 with
  | None => msg2
  | Some m => (fun _ => m (); msg2 ())
  end.

Ltac2 add_global_problem (prob : NamingProblems) (msg : unit -> unit) : unit :=
  prob.(global_problems) := Some (
    maybe_closure_append (prob.(global_problems)) msg
  ).

Ltac2 add_subgoal_problem (prob : NamingProblems) (i : int) (msg : unit -> unit) : unit :=
  prob.(subgoal_problems) := List.mapi (fun j om => 
    if Int.equal j i then
      Some (maybe_closure_append om msg)
    else
      om
  ) (prob.(subgoal_problems)).


Ltac2 reintroduce_with_pats (register_closure : (unit -> unit) -> unit) 
    (idents : ident list) (pats : Ltac1.t list) : unit :=
  let pat_len := List.length pats in
  let ident_len := List.length idents in
  let too_many_pats := Int.gt pat_len ident_len in
  let too_little_pats := Int.lt pat_len ident_len in
  let trunc_pat_list := (* cant use [to_list_of_size] since the introduction is staggered *)
    match too_many_pats with
    | true =>
      register_closure (fun _ => printf "Too many intro patterns were supplied!");
      register_closure (fun _ =>
        ltac1:(pats |- idtac "Dropping the following intro patterns:  " pats)
          (Ltac1.of_list (List.skipn ident_len pats))
      );
      firstn_safe (List.length idents) pats
    | false => pats
    end in

  let extra_reverts : ident list option ref := { contents := None } in
  Control.once (fun _ => 
    Control.plus (fun _ => Std.revert idents) (fun _ =>
      (* revert has failed, because of dependent hypothesis. 
        We compute all the dependent hyps, revert them for now, then reintroduce them later *)
      let i_deps := find_all_dependent_on_idents idents in
      extra_reverts.(contents) := Some i_deps ;
      Std.revert i_deps;
      Std.revert idents
    )
  );
  ltac1:(xs |- intros xs) (Ltac1.of_list trunc_pat_list);
  (* one might want to avoid [revert; intros] for a prettier proof term, 
     but [rename || destruct] does not work with i.e. intropattern [_] *)

  match too_little_pats with
  | false => ()
  | true =>
    let run_intros (_ : unit) := 
      (* We need to use [IntroFresh] here, since the provided patterns could conflict with the old names. *)
      Std.intros false (List.map (fun i => Std.IntroNaming (Std.IntroFresh i))
                       (List.skipn (List.length trunc_pat_list) idents))
    in
    register_closure (fun _ => printf "Too little intro patterns were supplied! Please supply a name for:");
    (* To obtain and print the actual names, we reuse [find_fresh_idents_between] *)
    let (f, cleanup) := find_fresh_idents_between () in
    run_intros ();
    let new_idents := f () in
    cleanup ();
    List.iter (fun i =>
      (* fprintf to print now, while focused, so that the closure still works when unfocused *)
      let msg := fprintf "  %I : %t" i (Constr.type (Control.hyp i)) in
      register_closure (fun _ => Message.print msg)
    ) new_idents
  end;

  (* Introduce the extra reverts, if there were any *)
  match extra_reverts.(contents) with
  | None => ()
  | Some is =>
    Std.intros false (List.map (fun i => Std.IntroNaming (Std.IntroIdentifier i)) is)
  end.

(* A weird interaction between the [unshelve] tactical and [clear] can
   result in pushing the [RecordCurrentVariables] evar to the current goal.
   This tactic will shelve all goals of that type *)
Ltac2 shelve_record_current_variables_goals (_ : unit) :=
  Control.enter (fun _ =>
    lazy_match! goal with
    | [ |- RecordCurrentVariables] => Control.shelve ()
    | [ |- _] => ()
    end
  ).

Ltac2 initialize_subgoal_problems (prob : NamingProblems) :=
  match prob.(subgoal_problems) with
  | [] =>
    prob.(subgoal_problems) := List.repeat None (count_goals ())
  | _ => ()
  end.

Ltac2 rename_multi_after_closure (prob : NamingProblems)
  (raw_multi_intro_pats : Ltac1.t)
  (closure : unit -> unit) :=
  let (f, cleanup) := find_fresh_idents_between () in
  closure ();
  shelve_record_current_variables_goals ();
  initialize_subgoal_problems prob;
  let raw_pat_list_list := Option.get (Ltac1.to_list raw_multi_intro_pats) in
  let pat_list_to_closure := (fun i pat_list =>
    (fun _ =>
      let new_idents := f () in
      let raw_pat_list := Option.get (Ltac1.to_list pat_list) in
      reintroduce_with_pats (add_subgoal_problem prob i) new_idents raw_pat_list
    )
  ) in
  let pat_list_list := to_list_of_size raw_pat_list_list (count_goals ()) (Ltac1.of_list [])
      (fun i => add_global_problem prob (fun _ => 
        printf "Too little branches of variable names supplied! Missing %i branches" i 
      ))
      (fun i => add_global_problem prob (fun _ =>
        printf "Too many branches of variable names supplied! Dropping %i branches" i
      ))
  in
  Control.dispatch (List.mapi pat_list_to_closure pat_list_list);
  cleanup ().

Ltac2 print_problems (prob : NamingProblems) : unit :=
  match prob.(global_problems) with
  | None => ()
  | Some c => c (); printf ""
  end;
  List.iteri (fun i mc =>
    match mc with
    | None => ()
    | Some c => 
      if Int.gt (List.length (prob.(subgoal_problems))) 1 then
        printf "Subgoal %i naming problems:" (Int.add i 1);
        c ();
        printf ""
      else c ()
    end
  ) (prob.(subgoal_problems)).

Ltac2 no_problems (_ : unit) : NamingProblems := {
  global_problems := None;
  subgoal_problems := [];
}.

Ltac zify_pre_hook ::=
  lazymatch goal with
  | f : utils_ltac2.RecordCurrentVariables |- _ =>
    clear f
  | |- _ => idtac
  end.



