From iris.proofmode Require Import proofmode.
From iris.algebra Require Import gmap.

From dislog.lang Require Export semantics.
From dislog.lang Require Export notation atomic.
From dislog.logic Require Import interp enc wpg_more monpred2 all_abef wpm.

Section group.
Context `{interpGS Σ}.

(******************************************************************************)
(* The whole machinery below is not needed if the locations does not
   points-to any location. *)

Lemma mapsto_at_unboxed l q vs :
  Forall (fun x => ¬ is_loc x) vs ->
  (l ↦{q} vs) ⊣⊢ embed (gen_heap.mapsto l q (SBlock vs)).
Proof.
  intros Hf. constructor. iIntros. rewrite mapsto_at monPred_at_embed.
  iSplit.
  { iIntros "(?&?)". iFrame. }
  { iIntros. iFrame. iInduction vs as [|v vs] "IH". by iApply all_abef_nil.
    iApply big_sepL_cons.
    apply Forall_cons in Hf. destruct Hf as (?&?).
    iSpecialize ("IH" with "[%//]"). iFrame "#". destruct v; done. }
Qed.

Implicit Type (M:list (loc * dfrac * list val)).

(******************************************************************************)

Definition group M : vProp Σ :=
  ([∗ list] x ∈ M, let '(l,q,vs) := x in l ↦{q} vs).

Global Instance group_timeless M : Timeless (group M).
Proof. apply big_sepL_timeless. intros ? ((?,?),?). apply _. Qed.

Lemma group_empty : ⊢ group nil.
Proof. done. Qed.
Lemma group_singleton l q vs : group [(l,q,vs)] ⊣⊢ (l ↦{q} vs)%I.
Proof. rewrite /group big_sepL_singleton //. Qed.
Lemma group_cons l q vs L : group ((l,q,vs) :: L) ⊣⊢ (l ↦{q} vs ∗ group L)%I.
Proof. rewrite /group big_sepL_cons //. Qed.
Lemma group_eq L :
  group L =  ([∗ list] x ∈ L, let '(l,q,vs) := x in l ↦{q} vs)%I.
Proof. done. Qed.

(******************************************************************************)

Definition third {A B C:Type} : (A*B*C) -> C := fun '(_,_,x) => x.
Definition values M : gset val := ⋃ ((fun x => list_to_set (third x)) <$> M).

Lemma values_empty : values nil = ∅.
Proof. set_solver. Qed.
Lemma values_singleton l q vs : values [(l,q,vs)] = list_to_set vs.
Proof. set_solver. Qed.

Lemma elem_of_values l M :
  l ∈ (values M) <->  exists i ls, (M !! i = Some ls /\ l ∈ third ls).
Proof.
  rewrite /values elem_of_union_list. split; intros Hi.
  { destruct Hi as (X,(H1&H2)).
    rewrite elem_of_list_fmap in H1. destruct H1 as (((?&?)&?)&?&?). subst.
    simpl in *. apply elem_of_list_to_set in H2.
    apply elem_of_list_lookup in H1. naive_solver. }
  { destruct Hi as  (i,(((?,?),?),(?&?))). simpl in *.
    exists (list_to_set l1). rewrite elem_of_list_to_set. split; last done.
    apply elem_of_list_lookup. exists i. rewrite list_lookup_fmap H0. done. }
Qed.

Lemma values_conv M :
  ([∗ set] v ∈ (values M), v ◷?now) ⊣⊢ ([∗ list] x ∈ M, [∗ list] v ∈ (third x), v ◷?now).
Proof.
  iSplit; iIntros "#HM"; rewrite big_sepS_forall big_sepL_forall.
  { iIntros (l ((?,?),?) Hl).
    iApply big_sepL_intro.
    iModIntro. iIntros (i v Hi).
    iApply "HM". iPureIntro. rewrite elem_of_values.
    eexists _,_. split; first done. apply elem_of_list_lookup. naive_solver. }
  { iIntros (v Hv).
    apply elem_of_values in Hv. destruct Hv as (?,(((?,?),?)&?&?)).
    iSpecialize ("HM" with "[%//]"). apply elem_of_list_lookup in H1.
    destruct H1. rewrite big_sepL_forall. by iApply "HM". }
Qed.

(******************************************************************************)

Lemma split_subj_obj (P:vProp Σ) :
  P ⊣⊢ (∃ t, monPred_in t ∗ embed (P t)).
Proof.
  iSplit; iIntros "H".
  { by iApply monPred_in_intro. }
  { iDestruct "H" as "[%t (?&?)]".
    iApply (monPred_in_elim with "[$][$]"). }
Qed.

Definition safe X : vProp Σ :=
  [∗ set] v ∈ X, v ◷?now.

Lemma group_get_clockm M :
  group M -∗ safe (values M).
Proof.
  iIntros. rewrite /group /safe.
  rewrite values_conv.
  iApply (big_sepL_mono with "[$]"). iIntros (i ((l,q),vs) Hl) "Hp".
  iDestruct "Hp" as "(?&?)". iFrame.
Qed.

Lemma punch_in M :
  group M -∗ ∃t, monPred_in t ∗ embed (group M t ∗ safe (values M) t).
Proof.
  iIntros.
  iDestruct (group_get_clockm with "[$]") as "#?".
  iDestruct (monPred_in_intro (group M ∗ safe (values M))%I with "[$]") as "[% (?&?)]".
  rewrite monPred_at_sep. iExists _. iFrame.
Qed.

Definition safe_upd (S S':gset val) :=
  forall v, v ∈ (S' ∖ S) -> ¬ (is_loc v).

Lemma safe_update S S' :
  safe_upd S S' ->
  safe S -∗
  safe S'.
Proof.
  iIntros (Hsafe) "#Habef".
  rewrite /safe !big_sepS_forall. iIntros (v Hv).
  destruct_decide (decide (v ∈ S)) as Hi.
  { by iApply "Habef". }
  { assert (¬ is_loc v) by (apply Hsafe; set_solver).
    iStopProof. constructor. iIntros (?) "_".
    rewrite /vclockm. destruct v; rewrite ?monPred_at_emp //. }
Qed.

Lemma safe_update_project S S' t :
  safe_upd S S' ->
  safe S t -∗
  safe S' t.
Proof.
  intros X.
  destruct (safe_update _ _ X) as [HX].
  rewrite -monPred_wand_force. rewrite -HX.
  rewrite monPred_at_emp //.
Qed.

Lemma objectivize M t :
  embed (safe (values M) t) -∗
  group M -∗
  embed (group M t).
Proof.
  iIntros "#HM ?".
  iDestruct (monPred_in_intro (group M) with "[$]") as "[% (_&?)]".
  iModIntro.
  rewrite /group /safe. rewrite !monPred_at_big_sepL monPred_at_big_sepS.
  iApply (big_sepL_impl with "[$]").
  iModIntro. iIntros (? ((?,?),?) ?) "HS".
  rewrite !mapsto_at. rewrite big_sepS_forall.
  iDestruct "HS" as "(?&_)". iFrame. rewrite /all_abef.
  rewrite big_sepL_forall. iIntros (???).
  destruct x; try done.
  iSpecialize ("HM" $! l1). simpl. iApply "HM". iPureIntro.
  apply elem_of_values. do 2 eexists. split; first done.
  by apply elem_of_list_lookup_2 in H1.
Qed.

Lemma objectivize' S M t :
  safe_upd S (values M) ->
  embed (safe S t) -∗
  group M -∗
  embed (group M t).
Proof.
  iIntros. iApply (objectivize with "[][$]"). iModIntro.
  iApply safe_update_project; done.
Qed.

Global Instance safe_upd_refl : Reflexive safe_upd.
Proof. intros ? ?. set_solver. Qed.

End group.

Global Opaque group.
